#include "ehs/io/mdl/Bone.h"

namespace ehs
{
	Bone::Bone()
		: hashName(0), id(0), animTrans(Mat4_f::Identity())
	{
	}

	Bone::Bone(Str_8 name, const UInt_8 id, const Mat4_f& localBindTrans, const Mat4_f& invBindTrans)
		: hashName(name.Hash_64()), name(std::move(name)), id(id), animTrans(Mat4_f::Identity()),
		localBindTrans(localBindTrans), invBindTrans(invBindTrans)
	{
	}

	Bone::Bone(Bone&& bone) noexcept
		: hashName(bone.hashName), name(std::move(bone.name)), id(bone.id), animTrans(bone.animTrans),
		  localBindTrans(bone.localBindTrans), invBindTrans(bone.invBindTrans), children(std::move(bone.children))
	{
		bone.hashName = 0;
		bone.id = 0xFF;
		bone.animTrans = Mat4_f::Identity();
		bone.localBindTrans = {};
		bone.invBindTrans = {};
	}

	Bone::Bone(const Bone& bone)
		: hashName(bone.hashName), name(bone.name), id(bone.id), animTrans(bone.animTrans), localBindTrans(bone.localBindTrans),
		invBindTrans(bone.invBindTrans), children(bone.children)
	{
	}

	Bone& Bone::operator=(Bone&& bone) noexcept
	{
		if (this == &bone)
			return *this;

		hashName = bone.hashName;
		name = std::move(bone.name);
		id = bone.id;
		animTrans = bone.animTrans;
		localBindTrans = bone.localBindTrans;
		invBindTrans = bone.invBindTrans;
		children = std::move(bone.children);

		bone.hashName = 0;
		bone.id = 0;
		bone.animTrans = Mat4_f::Identity();
		bone.localBindTrans = {};
		bone.invBindTrans = {};

		return *this;
	}

	Bone& Bone::operator=(const Bone& bone)
	{
		if (this == &bone)
			return *this;

		hashName = bone.hashName;
		name = bone.name;
		id = bone.id;
		animTrans = bone.animTrans;
		localBindTrans = bone.localBindTrans;
		invBindTrans = bone.invBindTrans;
		children = bone.children;

		return *this;
	}

	UInt_64 Bone::GetHashName() const
	{
		return hashName;
	}

	void Bone::SetName(Str_8 newId)
	{
		hashName = newId.Hash_64();
		name = std::move(newId);
	}

	Str_8 Bone::GetName() const
	{
		return id;
	}

	UInt_8 Bone::GetId() const
	{
		return id;
	}

	void Bone::SetAnimTrans(const Mat4_f& newTrans)
	{
		animTrans = newTrans;
	}

	Mat4_f Bone::GetAnimTrans() const
	{
		return animTrans;
	}

	void Bone::GetAnimTransRec(Array<Mat4_f>& output) const
	{
		output[id] = animTrans;

		for (UInt_64 i = 0; i < children.Size(); ++i)
			children[i].GetAnimTransRec(output);
	}

	Mat4_f Bone::GetLocalBindTrans() const
	{
		return localBindTrans;
	}

	Mat4_f Bone::GetInvBindTrans() const
	{
		return invBindTrans;
	}

	UInt_8 Bone::GetBoneCount() const
	{
		UInt_8 count = 1;

		for (UInt_64 i = 0; i < children.Size(); ++i)
			count += children[i].GetBoneCount();

		return count;
	}

	bool Bone::HasBone(const UInt_64 hashName, const UInt_8 id) const
	{
		if (this->hashName == hashName && this->id == id)
			return true;

		bool result = false;

		for (UInt_64 i = 0; i < children.Size(); ++i)
			result = children[i].HasBone(hashName, id);

		return result;
	}

	bool Bone::HasBone(const UInt_64 hashName) const
	{
		if (this->hashName == hashName)
			return true;

		bool result = false;

		for (UInt_64 i = 0; i < children.Size(); ++i)
			result = children[i].HasBone(hashName);

		return result;
	}

	bool Bone::HasBone(const UInt_8 id) const
	{
		if (this->id == id)
			return true;

		bool result = false;

		for (UInt_64 i = 0; i < children.Size(); ++i)
			result = children[i].HasBone(id);

		return result;
	}

	bool Bone::AddBone(Bone child)
	{
		if (HasBone(child.hashName, child.id))
			return false;

		children.Push(std::move(child));

		return true;
	}

	const Bone* Bone::GetBone(const UInt_64 hashName) const
	{
		if (this->hashName == hashName)
			return this;

		const Bone* result = nullptr;

		for (UInt_64 i = 0; i < children.Size(); ++i)
			result = children[i].GetBone(hashName);

		return result;
	}

	Bone* Bone::GetBone(const UInt_64 hashName)
	{
		if (this->hashName == hashName)
			return this;

		Bone* result = nullptr;

		for (UInt_64 i = 0; i < children.Size(); ++i)
			result = children[i].GetBone(hashName);

		return result;
	}

	const Bone* Bone::GetBone(const UInt_8 id) const
	{
		if (this->id == id)
			return this;

		const Bone* result = nullptr;

		for (UInt_64 i = 0; i < children.Size(); ++i)
			result = children[i].GetBone(id);

		return result;
	}

	Bone* Bone::GetBone(const UInt_8 id)
	{
		if (this->id == id)
			return this;

		Bone* result = nullptr;

		for (UInt_64 i = 0; i < children.Size(); ++i)
			result = children[i].GetBone(id);

		return result;
	}

	const Array<Bone>& Bone::GetChildren() const
	{
		return children;
	}

	Array<Bone>& Bone::GetChildren()
	{
		return children;
	}
}