#include "ehs/json/Json.h"
#include "ehs/Log.h"

namespace ehs
{
    Json::~Json()
    {
        delete value;
    }

    Json::Json()
        : value(nullptr)
    {
    }

    Json::Json(const JsonBase& value)
        : value(new JsonBase(value))
    {
    }

    Json::Json(const JsonObj& value)
        : value((JsonBase*)new JsonObj(value))
    {
    }

    Json::Json(const JsonArray& value)
        : value((JsonBase*)new JsonArray(value))
    {
    }

    Json::Json(const JsonBool& value)
        : value((JsonBase*)new JsonBool(value))
    {
    }

    Json::Json(const JsonNum& value)
            : value((JsonBase*)new JsonNum(value))
    {
    }

    Json::Json(const JsonStr& value)
            : value((JsonBase*)new JsonStr(value))
    {
    }
	
	Json::Json(const char* data, const UInt_64 size, const UInt_64 extra)
        : value(nullptr)
	{
        Parse({data, size}, extra);
	}
	
	Json::Json(const Str_8& data, const UInt_64 extra)
        : value(nullptr)
	{
        Parse(data, extra);
	}

	Json::Json(Json&& json) noexcept
		: value(json.value)
	{
		json.value = nullptr;
	}

    Json::Json(const Json& json)
        : value(nullptr)
    {
        if (!json.value)
            return;

        switch (json.value->GetType())
        {
            case JsonType::OBJ:
                value = (JsonBase*)new JsonObj(*(JsonObj*)json.value);
                break;
            case JsonType::ARRAY:
                value = (JsonBase*)new JsonArray(*(JsonArray*)json.value);
                break;
            case JsonType::BOOL:
                value = (JsonBase*)new JsonBool(*(JsonBool*)json.value);
                break;
            case JsonType::NUM:
                value = (JsonBase*)new JsonNum(*(JsonNum*)json.value);
                break;
            case JsonType::STR:
                value = (JsonBase*)new JsonStr(*(JsonStr*)json.value);
                break;
            default:
                value = new JsonBase();
                break;
        }
    }

	Json& Json::operator=(Json&& json) noexcept
	{
		if (this == &json)
			return *this;

		delete value;
		value = json.value;

		json.value = nullptr;

		return *this;
	}

    Json& Json::operator=(const Json& json)
    {
        if (this == &json)
            return *this;

        delete value;
        value = nullptr;

        if (!json.value)
            return *this;

        switch (json.value->GetType())
        {
            case JsonType::OBJ:
                value = (JsonBase*)new JsonObj(*(JsonObj*)json.value);
                break;
            case JsonType::ARRAY:
                value = (JsonBase*)new JsonArray(*(JsonArray*)json.value);
                break;
            case JsonType::BOOL:
                value = (JsonBase*)new JsonBool(*(JsonBool*)json.value);
                break;
            case JsonType::NUM:
                value = (JsonBase*)new JsonNum(*(JsonNum*)json.value);
                break;
            case JsonType::STR:
                value = (JsonBase*)new JsonStr(*(JsonStr*)json.value);
                break;
            default:
                value = new JsonBase();
                break;
        }

        return *this;
    }

    JsonBase* Json::GetValue()
    {
        return value;
    }

    JsonBase* Json::RetrieveValue(const Str_8& access)
    {
        if (!value || !access.Size())
            return nullptr;

        Vector<Str_8> levels = ParseAccess(access);

        JsonBase* hierarchy = value;

        for (UInt_64 i = 0; i < levels.Size(); ++i)
        {
            if (hierarchy->GetType() == JsonType::ARRAY)
            {
                if (levels[i][0] >= '0' && levels[i][0] <= '9')
                {
                    hierarchy = (*((JsonArray*)hierarchy))[levels[i].ToDecimal<UInt_64>()];
                }
                else
                    return nullptr;
            }
            else if (hierarchy->GetType() == JsonType::OBJ)
            {
                if (levels[i][0] < '0' || levels[i][0] > '9')
                {
                    JsonVar* var = ((JsonObj*)hierarchy)->GetVar(levels[i]);
                    if (!var || !var->GetValue())
                        return nullptr;

                    hierarchy = var->GetValue();
                }
                else
                    return nullptr;
            }
            else
                return nullptr;
        }

        return hierarchy;
    }

    Str_8 Json::ToStr(const bool compact) const
    {
        return value->ToStr(0, compact);
    }

    Vector<Str_8> Json::ParseAccess(const Str_8& access)
    {
        bool isIndex = false;
        const Char_8* start = &access[0];
        const Char_8* end = &access[access.Size() - 1];

        Vector<Str_8> levels(0, 5);

        for (const Char_8* i = &access[0]; i <= end; ++i)
        {
            if (*i == '[')
            {
                if (i - start)
                    levels.Push(Str_8(start, i - start));

                start = i + 1;
                isIndex = true;
            }
            else if (*i == ']')
            {
                if (!isIndex)
                {
                    EHS_LOG_INT(LogType::ERR, 0, "Index has ended, but never started.");
                    return levels;
                }

                if (i - start)
                    levels.Push(Str_8(start, i - start));
                else
                {
                    EHS_LOG_INT(LogType::ERR, 1, "Index has no value.");
                    return levels;
                }

                start = i + 1;
                isIndex = false;
            }
            else if (isIndex && (*i < '0' || *i > '9'))
            {
                EHS_LOG_INT(LogType::ERR, 2, "Index has an invalid character, \"" + Str_8(i, 1) + "\". Index must be a whole number.");
                return levels;
            }
            else if (*i == '.')
            {
                if (i - start)
                    levels.Push(Str_8(start, i - start));

                start = i + 1;
            }
            else if (!isIndex && *i != '-' && *i != '_' && (*i < 'A' || *i > 'Z') && (*i < 'a' || *i > 'z'))
            {
                EHS_LOG_INT(LogType::ERR, 3, "Member variable has an invalid character, \"" + Str_8(i, 1) + "\".");
                return levels;
            }
        }

        if (end + 1 - start)
            levels.Push(Str_8(start, end + 1 - start));

        return levels;
    }

    void Json::ParseValue(JsonVar* var, const Char_8** begin, const Char_8* end, const UInt_64 extra)
    {
        const Char_8* start = nullptr;
		bool isStr = false;

        for (; *begin <= end; ++(*begin))
        {
            if (**begin == '\"')
            {
                if (start)
                {
                    if (*begin - start)
						((JsonStr*)var->GetValue())->value = Str_8(start, *begin - start);

                    ++(*begin);
                    return;
                }
                else
				{
					start = *begin + 1;
					isStr = true;
					var->SetValue(JsonStr());
				}
            }
            else if (!start && **begin == '{')
            {
                ++(*begin);
                JsonObj obj(extra);
                ParseObject(&obj, begin, end, extra);
                var->SetValue(obj);
                return;
            }
            else if (!start && **begin == '[')
            {
                ++(*begin);
                JsonArray arr(extra);
                ParseArray(&arr, begin, end, extra);
                var->SetValue(arr);
                return;
            }
            else if (!start && **begin != ' ' && **begin != ',' && **begin != '\r' && **begin != '\n' && **begin != '\t')
            {
                start = *begin;
            }
            else if (start && !isStr && (**begin == '\t' || **begin == '\r' || **begin == '\n' || (!var->GetValue() && **begin == ' ') || **begin == ',' || **begin == '}' || **begin == ']'))
            {
                Str_8 result(start, *begin - start);
                if (result == "true")
                    var->SetValue(true);
                else if (result == "false")
                    var->SetValue(false);
                else if (result == "null")
                    var->SetValue(JsonBase());
                else if (result.IsNum())
                    var->SetValue(result.ToFloat());

                return;
            }
        }
    }

    JsonVar Json::ParseVar(const Char_8** begin, const Char_8* end, const UInt_64 extra)
    {
        JsonVar var;

        const Char_8* start = nullptr;

        for (; *begin <= end; ++(*begin))
        {
            if (**begin == '}' || **begin == ']')
            {
                ++(*begin);
                return var;
            }
            else if (**begin == '"')
            {
                if (start)
                {
                    var = JsonVar(Str_8(start, *begin - start));
                    start = nullptr;
                }
                else
                    start = *begin + 1;
            }
            else if (!start && **begin == ':')
            {
                if (var.GetHashId())
                {
                    ++(*begin);

                    ParseValue(&var, begin, end, extra);

                    return var;
                }
            }
        }

        return var;
    }

    void Json::ParseObject(JsonObj* obj, const Char_8** begin, const Char_8* end, const UInt_64 extra)
    {
        JsonVar var;
        do
        {
            var = ParseVar(begin, end, extra);
        }
        while(obj->AddVar(var));
    }

    void Json::ParseArray(JsonArray* arr, const Char_8** begin, const Char_8* end, const UInt_64 extra)
    {
        const Char_8* start = nullptr;

        for (; *begin <= end; ++(*begin))
        {
            if (**begin == '}' || **begin == ']')
            {
                ++(*begin);
                return;
            }
            else if (**begin == '\"')
            {
                if (start)
                {
                    if (*begin - start)
                        arr->Push(start, *begin - start);

					start = nullptr;
                    ++(*begin);
                    continue;
                }
                else
                    start = *begin + 1;
            }
            else if (!start && **begin == '{')
            {
                ++(*begin);
                JsonObj obj(extra);
                ParseObject(&obj, begin, end, extra);
                arr->Push(obj);
                continue;
            }
            else if (!start && **begin == '[')
            {
                ++(*begin);
                JsonArray array(extra);
                ParseArray(&array, begin, end, extra);
                arr->Push(array);
                continue;
            }
            else if (!start && **begin != ' ' && **begin != ',' && **begin != '\r' && **begin != '\n' && **begin != '\t')
            {
                start = *begin;
            }
            else if (start && *start != '\"' && (**begin == '\t' || **begin == '\r' || **begin == '\n' || **begin == ' ' || **begin == ',' || **begin == '}' || **begin == ']'))
            {
                Str_8 result(start, *begin - start);
                if (result == "true")
                    arr->Push(true);
                else if (result == "false")
                    arr->Push(false);
                else if (result == "null")
                    arr->Push(JsonBase());
                else if (result.IsNum())
                    arr->Push(result.ToFloat());

                continue;
            }
        }
    }

    void Json::Parse(const Str_8& data, const UInt_64 extra)
    {
        const Char_8* start = nullptr;
        const Char_8* end = &data[data.Size() - 1];

        for (const Char_8* i = &data[0]; i <= end; ++i)
        {
            if (*i == '}' || *i == ']')
            {
                ++i;
                return;
            }
            else if (*i == '\"')
            {
                if (start)
                {
                    if (i - start)
                        value = new JsonStr(start, i - start);

                    ++i;
                    return;
                }
                else
                    start = i + 1;
            }
            else if (!start && *i == '{')
            {
                ++i;
                JsonObj* obj = new JsonObj(extra);
                ParseObject(obj, &i, end, extra);
                value = obj;
                return;
            }
            else if (!start && *i == '[')
            {
                ++i;
                JsonArray* arr = new JsonArray(extra);
                ParseArray(arr, &i, end, extra);
                value = arr;
                return;
            }
            else if (!start && *i != ' ' && *i != ',' && *i != '\r' && *i != '\n' && *i != '\t')
            {
                start = i;
            }
            else if (start && *start != '\"' && (*i == '\t' || *i == '\r' || *i == '\n' || *i == ' ' || *i == ',' || *i == '}' || *i == ']'))
            {
                Str_8 result(start, i - start);
                if (result == "true")
                    value = new JsonBool(true);
                else if (result == "false")
                    value = new JsonBool(false);
                else if (result == "null")
                    value = new JsonBase();
                else if (result.IsNum())
                    value = new JsonNum(result.ToFloat());

                return;
            }
        }
    }
}